################################################################################
# author: Jona Cederbaum
################################################################################
# description: function to generate simulation data 
# is called by gendata_crossed
################################################################################
gendata_fun  <-  function(nrep, Iseq, Jseq, nu_B_fun, nu_C_fun, nu_E_fun, 
                          sigmasq, normal, mu, N_B, N_C, N_E, phi_B_fun, 
                          phi_C_fun, phi_E_fun, minsize, maxsize, min_grid, max_grid, 
                          min_visit, max_visit, use_RI_true, use_simple_true, 
                          center_basis_weights, decor_basis_weights){
  
  #####################
  # checks and warnings
  #####################
  if(nrep < 1)
    stop("the number of iterations must be larger than 0")
  
  if(is.na(Iseq))
    stop("Iseq needs to be not NA")
  
  if(Iseq == 0)
    stop("Iseq needs to be larger than 0")
  
  if(!use_RI_true & use_simple_true)
    stop("when using use_simple_true == TRUE, please also set use_RI_true to TRUE")
  
  if(use_RI_true & !is.na(Jseq))
    stop("when using use_RI_true, Jseq needs to be set to 'NA'")
  
  if(use_RI_true & !is.na(nu_C_fun))
    stop("when using use_RI_true, nu_C_fun needs to be set to 'NA'")
  
  if(use_RI_true & !is.na(N_C))
    stop("when using use_RI_true, N_C needs to be set to 'NA'")
  
  if(use_RI_true & !is.na(phi_C_fun))
    stop("when using use_RI_true, phi_C_fun needs to be set to 'NA'")
  
  if(use_RI_true & !is.na(phi_C_fun))
    stop("when using use_RI_true, phi_C_fun needs to be set to 'NA'")
  
  if(use_simple_true & !is.na(nu_E_fun))
    stop("when using use_simple_true, nu_E_fun needs to be set to 'NA'")
  
  if(use_simple_true & (N_E != 0))
    stop("when using use_simple_true, N_E needs to be set to '0'")
  
  if(use_simple_true & !is.na(phi_E_fun))
    stop("when using use_simple_true, phi_E_fun needs to be set to 'NA'")
  
  if(min_grid != 0 | max_grid != 1)
    stop("so far, the data generating process is only implemented for measurement points in [0,1]")
  
  if(!use_RI_true & is.na(Jseq))
    stop("for crossed fRIs (use_RI_true == FALSE), please specify Jseq other than NA")
  
  if(!use_RI_true & is.na(N_B))
    stop("for crossed fRIs (use_RI_true == FALSE), please specify N_B other than NA")
  
  if(!use_RI_true & N_B == 0)
    stop("for crossed fRIs (use_RI_true == FALSE), please specify N_B other than 0")
  
  if(!use_RI_true & is.na(N_C))
    stop("for crossed fRIs (use_RI_true == FALSE), please specify N_C other than NA")
  
  if(!use_RI_true & N_C == 0)
    stop("for crossed fRIs (use_RI_true == FALSE), please specify N_C other than 0")
  
  if(!use_RI_true & is.na(N_E))
    stop("for crossed fRIs (use_RI_true == FALSE), please specify N_E other than NA")
  
  if(!use_RI_true & N_E == 0)
    stop("for crossed fRIs (use_RI_true == FALSE), please specify N_E other than 0")
  
  if(!use_RI_true & is.na(phi_B_fun))
    stop("for crossed fRIs (use_RI_true == FALSE), please specify phi_B_fun other than NA")
  
  if(!use_RI_true & is.na(phi_C_fun))
    stop("for crossed fRIs (use_RI_true == FALSE), please specify phi_C_fun other than NA")
  
  if(!use_RI_true & is.na(phi_E_fun))
    stop("for crossed fRIs (use_RI_true == FALSE), please specify phi_E_fun other than NA")
  
  if(!use_RI_true & is.na(nu_B_fun))
    stop("for crossed fRIs (use_RI_true == FALSE), please specify nu_B_fun other than NA")
  
  if(!use_RI_true & is.na(nu_C_fun))
    stop("for crossed fRIs (use_RI_true == FALSE), please specify nu_C_fun other than NA")
  
  if(!use_RI_true & is.na(nu_E_fun))
    stop("for crossed fRIs (use_RI_true == FALSE), please specify nu_E_fun other than NA")
  
  ###################
  # initialize output
  ###################
  res  <-  list() 
  
  if(!use_RI_true){
    ##################
    # for crossed fRIs
    print("you are generating data for a two crossed fRIs")
  }else{
    if(!use_simple_true){
      ##############
      # for one fRIs
      print("you are generating data for a one fRI")  
    }else{
      ########################
      # for independent curves
      print("you are generating data for independent curves")  
    }
    Jseq  <-  NA
  }
  
  ########################################################################
  # simulation run for
  # one iteration
  ########################################################################
  for (r in 1:nrep){  
    ########################
    # iterate over levels of 
    # the grouping variables
    ########################
    for(I in Iseq){
      for(J in Jseq){
        cat("generate iteration r: ", r, "; I: ", I, "; J: ", J, "\n",  sep = "")
        
        if(!use_RI_true){
          if(min_visit != max_visit){
            Hvec  <-  sort(sample(x = (min_visit:max_visit), size = I * J, replace = TRUE, 
                                  prob = sample(x = 2:5, size = (max_visit - min_visit + 1), 
                                                replace = TRUE))) 
          }else{
            Hvec <- rep(min_visit, length = I * J)
          }
        }else{
          if(min_visit != max_visit){
            Hvec <- sort(sample(x = (min_visit:max_visit), size = I, replace = TRUE, 
                                prob = sample(x = 2:5, size = (max_visit - min_visit + 1), 
                                              replace = TRUE))) 
          }else{
            Hvec <- rep(min_visit, length = I)
          }
        }
        
        ############################
        # get total number of curves
        ############################
        n <- sum(Hvec)
        
        ##########################################
        # for crossed fRIs, take all combinations 
        # of first and second grouping variable
        ##########################################
        if(!use_RI_true){
          help <- expand.grid(1:J, 1:I)
          subject_help <- help$Var2
          word_help <- help$Var1
          subject <- rep(subject_help, Hvec)
          word <- rep(word_help, Hvec)
          combi_help <- list()
          for(i in 1:(I * J)){
            combi_help[[i]] <- 1:Hvec[i]
          }
          combi <- unlist(combi_help)
        }else{
          subject <- rep(1:I, times = Hvec)
        }
        
        
        ###################################################
        # for each curve, draw number of observation points
        ###################################################
        if(minsize != maxsize){
          number <- sample(x = rep(minsize:maxsize), size = n, replace = TRUE)   
        }else{
          number <- rep(minsize, length = n)
        }
        
        ####################################
        # make long data.table with
        # one row for each observation point
        ####################################
        number_long <- rep(number, number)
        n_long <- rep(1:n, number) 
        subject_long <- rep(subject, number)
        if(!use_RI_true){
          word_long <- rep(word, number)
          combi_long <- rep(combi, number)
          curve_info <- data.table(n_long, subject_long, word_long, combi_long, number_long)
        }else{
          curve_info <- data.table(n_long, subject_long, number_long)  
        }
        
        ################################################
        # for each curve, draw location of 
        # observation points from a uniform distribution
        ################################################
        locs <- list()
        for(i in 1:n){
          locs[[i]] <- runif(n = number[i], min = min_grid, max = max_grid)
        }
        
        #########################
        # sort measurement points
        #########################
        t <- list()
        o <- lapply(locs, order)
        for(i in 1:n) {
          help <- o[[i]]
          t[[i]] <- locs[[i]][help]}
        t_vec <- unlist(t)
        
        ########################
        # add measurement points
        # to data table
        ########################
        curve_info[, t := t_vec]
        
        
        ########################################
        # draw random basis weights of normal or 
        # mixure of normal distribution(s)
        ########################################
        if(N_B > 0){
          nu_B <- nu_B_fun(1:N_B)  
        }else{
          nu_B <- 0
          warning("N_B is not positive")
        }        
        if(!use_RI_true){
          ##################
          # for crossed fRIs
          if(N_C > 0){
            nu_C <- nu_C_fun(1:N_C)  
          }else{
            nu_C <- 0
          }
        }
        if(N_E > 0){
          #############
          # for one fRI
          nu_E <- nu_E_fun(1:N_E)  
        }else{
          ########################
          # for independent curves
          nu_E <- 0
          if(!use_simple_true)
            warning("N_E is not positive")
        }     
        
        ############################
        # for Gaussian basis weights
        if(normal == TRUE){
          if(N_B > 0){
            xi_B <- rmvnorm(I, mean = rep(0, N_B), sigma = diag(N_B) * nu_B)
          }else{
            xi_B <- rep(NA, length = I)
          }
          if(!use_RI_true){
            if(N_C > 0){
              xi_C <- rmvnorm(J, mean = rep(0, N_C), sigma = diag(N_C) * nu_C)
            }else{
              xi_C <- rep(NA, length = J)
            }
          }
          if(N_E > 0){
            xi_E <- rmvnorm(n, mean = rep(0, N_E), sigma = diag(N_E) * nu_E)   
          }else{
            xi_E <- rep(NA, length = n)
          }
        } 
        
        ################################
        # for non Gaussian basis weights
        if(normal == FALSE){ 
          if(N_B > 1){
            xi_B <- matrix(rnorm(I * N_B), I, N_B) %*% diag(sqrt(nu_B / 2)) + 
              matrix(2 * rbinom(I * N_B, 1, 0.5) - 1, I, N_B) %*% diag(sqrt(nu_B / 2))
          }else{
            if(N_B > 0){
              xi_B <- matrix(rnorm(I * N_B), I, N_B) %*% sqrt(nu_B/2) + 
                matrix(2 * rbinom(I * N_B, 1, 0.5) - 1, I, N_B) %*% sqrt(nu_B / 2)  
            } else{
              xi_B <- rep(NA, length = I)
            }
          }
          
          if(!use_RI_true){
            ##################
            # for crossed fRIs
            if(N_C > 1){
              xi_C <- matrix(rnorm(J * N_C), J, N_C)% * %diag(sqrt(nu_C / 2)) + 
                matrix(2 * rbinom(J * N_C, 1, 0.5) - 1, J, N_C) %*% diag(sqrt(nu_C / 2))
            }else{
              if(N_C > 0){
                xi_C <- matrix(rnorm(J * N_C), J, N_C) %*% sqrt(nu_C / 2) + 
                  matrix(2 * rbinom(J * N_C, 1, 0.5) - 1, J, N_C) %*% sqrt(nu_C / 2)  
              }else{
                xi_C <- rep(NA, length = J) 
              }
            }
          }
          if(N_E > 1){
            xi_E <- matrix(rnorm(n * N_E), n, N_E) %*% diag(sqrt(nu_E / 2)) + 
              matrix(2 * rbinom(n * N_E, 1, 0.5) - 1, n, N_E) %*% diag(sqrt(nu_E / 2))  
          }else{
            if(N_E > 0){
              xi_E <- matrix(rnorm(n * N_E), n, N_E) %*% sqrt(nu_E / 2) +
                matrix(2 * rbinom(n * N_E, 1, 0.5) - 1, n, N_E) %*% sqrt(nu_E / 2)  
            }else{
              xi_E <- rep(NA, length = n) 
            }
          }
        }
        
        ################################
        # center empirical basis weights
        ################################
        if(center_basis_weights){
          if(N_B > 0){
            for(k in 1:N_B){
              xi_B[, k] <- xi_B[, k] - mean(xi_B[, k])
            }
          }
          if(!use_RI_true){
            ##################
            # for crossed fRIs
            if(N_C > 0){
              for(k in 1:N_C){
                xi_C[, k] <- xi_C[, k] - mean(xi_C[, k])
              }
            }
          }
          if(N_E > 0){
            for(k in 1:N_E){
              xi_E[, k] <- xi_E[, k] - mean(xi_E[, k])
            }
          }
        }
        
        #####################################
        # decorrelate empirical basis weights
        # and correct empirical variance
        #####################################
        if(decor_basis_weights){
          if(N_B > 1)
            xi_B <- xi_B %*% sqrtm(solve(cov(xi_B))) %*% sqrtm(diag(nu_B, N_B, N_B))
          if(!use_RI_true){
            if(N_C > 1)
              xi_C <- xi_C % * %sqrtm(solve(cov(xi_C))) %*% sqrtm(diag(nu_B, N_C, N_C))
          }
          if(N_E > 1)
            xi_E <- xi_E %*% sqrtm(solve(cov(xi_E))) %*% sqrtm(diag(nu_B, N_E, N_E))
        }
        
        ########################################
        # repeat basis weights and to data table 
        ########################################
        if(N_B > 0){
          xi_B_long_help <- list()
          xi_B_long_help2 <- list()
          for(k in 1:N_B){
            xi_B_long_help[[k]] <- xi_B[, k][subject]
            xi_B_long_help2[[k]] <- rep(xi_B_long_help[[k]], number)
          }
          xi_B_long <- xi_B_long_help2[[1]]
          if(N_B > 1){
            for(k in 2:N_B){
              xi_B_long <- cbind(xi_B_long, xi_B_long_help2[[k]])
            }
          }else{
            xi_B_long <- matrix(xi_B_long, ncol = 1) 
          }
          for(k in 1:N_B){
            curve_info[, paste0("xi_B_long.", k) := xi_B_long[, k]]
          }  
        }
        if(!use_RI_true){
          ##################
          # for crossed fRIs
          if(N_C > 0){
            xi_C_long_help <- list()
            xi_C_long_help2 <- list()
            for(k in 1:N_C){
              xi_C_long_help[[k]] <- xi_C[, k][word]             
              xi_C_long_help2[[k]] <- rep(xi_C_long_help[[k]], number)
            }
            xi_C_long <- xi_C_long_help2[[1]]
            if(N_C > 1){
              for(k in 2:N_C){
                xi_C_long <- cbind(xi_C_long, xi_C_long_help2[[k]])
              }
            }else{
              xi_C_long <- matrix(xi_C_long, ncol = 1) 
            }
            for(k in 1:N_C){
              curve_info[, paste0("xi_C_long.", k) := xi_C_long[, k]]
            }  
          }          
        } 
        if(N_E > 0){
          xi_E_long_help <- list()
          for(k in 1:N_E){
            xi_E_long_help[[k]] <- rep(xi_E[, k], number)  
          }
          xi_E_long <- xi_E_long_help[[1]]
          if(N_E > 1){
            for(k in 2:N_E){
              xi_E_long <- cbind(xi_E_long, xi_E_long_help[[k]])
            }  
          }else{
            xi_E_long <- matrix(xi_E_long, ncol = 1) 
          }
          for(k in 1:N_E){
            curve_info[, paste0("xi_E_long.", k) := xi_E_long[, k]]
          }  
        }
        
        ######################################################################
        # construct true underlying random functions B_i(t), C_i(t) and E_i(t)
        ######################################################################
        if(N_B > 0){
          phi_B <- sapply(1:N_B, phi_B_fun, t = t_vec)
        }else{
          phi_B <- rep(NA, length = length(t_vec))
        }
        if(!use_RI_true){
          ##################
          # for crossed fRIs
          if(N_C > 0){
            phi_C <- sapply(1:N_C, phi_C_fun, t = t_vec)
          }else{
            phi_C <- rep(NA, length = length(t_vec))
          }
        }
        if(N_E > 0){
          phi_E <- sapply(1:N_E, phi_E_fun, t = t_vec)
        }else{
          phi_E <- rep(NA, length = length(t_vec))
        }
        if(N_B > 0){
          B_vec_help <- matrix(NA, ncol = N_B, nrow = length(n_long))
          for(k in 1:N_B){
            B_vec_help[, k] <- curve_info[[paste("xi_B_long.", k, sep = "")]] * phi_B[, k]
          }
          B_vec <- rowSums(B_vec_help)        
        }else{
          B_vec <- rep(0, length = length(t_vec))
        }
        if(!use_RI_true){
          ##################
          # for crossed fRIs
          if(N_C > 0){
            C_vec_help <- matrix(NA, ncol = N_C, nrow = length(n_long))
            for(k in 1:N_C){
              C_vec_help[, k] <- curve_info[[paste("xi_C_long.", k, sep = "")]] * phi_C[, k]
            }       
            C_vec <- rowSums(C_vec_help)        
          }else{
            C_vec <- rep(0, length = length(t_vec))
          }
        }
        if(N_E > 0){
          E_vec_help <- matrix(NA, ncol = N_E, nrow = length(n_long))  
          for(k in 1:N_E){
            E_vec_help[, k] <- curve_info[[paste("xi_E_long.", k, sep = "")]] * phi_E[, k]
          }        
          E_vec <- rowSums(E_vec_help)
        }else{
          E_vec <- rep(0, length = length(t_vec))
        }
        if(N_B > 0){
          curve_info[, "B_long" := B_vec]
        }
        if(!use_RI_true){
          if(N_C > 0){
            curve_info[, "C_long" := C_vec]
          }
        }
        if(N_E > 0){
          curve_info[, "U_long" := E_vec]
        }
        
        ###########################################
        # draw random white noise measurement error
        # from a multivariate normal distribution
        ###########################################
        epsilon <- list()
        for(i in 1:n){
          if(number[i] == 1){
            epsilon[[i]] <- rmvnorm(1, mean = rep(0, number[i]), 
                                    sigma = as.matrix(sigmasq, nrow = 1, ncol = 1))  
          }else{
            epsilon[[i]] <- rmvnorm(1, mean = rep(0, number[i]), 
                                    sigma = diag(rep(sigmasq, length = number[i])))  
          }
        }
        epsilon_vec <- unlist(epsilon)
        
        #########################
        # construct mean function
        #########################
        mu_list <- list()
        for(i in 1:n){
          t_use <- subset(curve_info, select = t, n_long == i)
          mu_list[[i]] <- mu(t_use)
        }   
        mu_vec <- unlist(mu_list)
        
        
        ###################################
        # construct functional observations
        ###################################
        if(!use_RI_true){
          ##################
          # for crossed fRIs
          y_vec <- mu_vec + B_vec + C_vec + E_vec + epsilon_vec   
        }else{
          y_vec <- mu_vec + B_vec + E_vec + epsilon_vec  
        }
        
        #################################
        # add mean, measurement error and
        # observations to data table
        #################################
        curve_info[, c("mu_vec", "epsilon_vec", "y_vec") := list(mu_vec, epsilon_vec, y_vec)]
        
        ###############
        # reutrn output
        ###############
        res[["nu_B_true"]] <- nu_B
        if(!use_RI_true)
          res[["nu_C_true"]] <- nu_C  
        res[["nu_E_true"]] <- nu_E
        res[["mu_true"]] <- mu
        res[["phi_B_true"]] <- phi_B
        if(!use_RI_true)
          res[["phi_C_true"]] <- phi_C
        res[["phi_E_true"]] <- phi_E
        res[["phi_B_fun_true"]] <- phi_B_fun
        if(!use_RI_true)
          res[["phi_C_fun_true"]] <- phi_C_fun
        res[["phi_E_fun_true"]] <- phi_E_fun
        res[["minsize_true"]] <- minsize
        res[["maxsize_true"]] <- maxsize
        res[["min_grid_true"]] <- min_grid
        res[["max_grid_true"]] <- max_grid
        res[["min_visit_true"]] <- min_visit
        res[["max_visit_true"]] <- max_visit
        res[["sigmasq_true"]] <- sigmasq
        res[["nrep"]] <- nrep
        res[["Iseq"]] <- Iseq
        res[["Jseq"]] <- Jseq
        res[[paste("number_true_I", I, "_J", J, "_r", r, sep = "")]] <- number
        res[[paste("curve_info_true_I", I, "_J", J, "_r", r, sep = "")]] <- curve_info
        res[["normal"]] <- normal
        res[[paste0(" xi_B_true_I", I, "_J", J, "_r", r)]] <- xi_B 
        if(!use_RI_true)
          res[[paste0(" xi_C_true_I", I, "_J", J, "_r", r)]] <- xi_C
        res[[paste0(" xi_E_true_I", I, "_J", J, "_r", r)]] <- xi_E
        res[["center_basis_weights"]] <- center_basis_weights
        res[["decor_basis_weights"]] <- decor_basis_weights
        
      }
    }
  }
  return(res)
}

################################################################################
