################################################################################
# author: Jona Cederbaum
################################################################################
# description: main frame for simulation after data have been generated
# calls function in call_all_functions.R.
################################################################################
sim_fun <- function(bf_mean, bf_covariates, bf_covs, method, m_mean, m_covs, bs, 
                    d_grid, var_level, N_B, N_C, N_E, use_bam, covariate, covariate_form, 
                    num_covariates, dat_gen, interaction, which_interaction, 
                    use_famm, mp, para_estim, para_estim_nc, use_tri, use_whole, 
                    use_tri_constr, use_tri_constr_weights, use_bam_famm, 
                    use_RI,  para_estim_famm, para_estim_famm_nc, 
                    save_model_mean, family = "normal", bs_int_famm, bs_y_famm, 
                    use_simple, np, min_grid, max_grid, my_grid, save_model_famm, nrep_use){
  
  for(f in family){
    ###################
    # load simulation
    # data for family f
    ###################
    if(!use_RI){
      ##################
      # for crossed fRIs
      load(paste0("sim_data/gen_", f, "_crossed", dat_gen, ".Rdata"))
    }else{
      if(!use_simple){
        #############
        # for one fRI
        load(paste0("sim_data/gen_", f, "_RI", dat_gen, ".Rdata"))  
      }else{
        ########################
        # for independent curves
        load(paste0("sim_data/gen_", f, "_simple", dat_gen, ".Rdata"))  
      }
    }
    
    #########################
    # reduce iteration number
    # if desired
    #########################
    if(!is.null(nrep_use)){
      if(nrep_use <= nrep){
        nrep <- nrep_use
        warning(paste0("nrep is reduced to ", nrep))
      }else{
        warning(paste0("nrep is chosen too be higher than the existing ", nrep, " data sets "))
      }
    }
    
    ###########################
    # generate name for results
    # depending on the date
    # and type of scenario
    ###########################
    if(!use_RI){
      ##################
      # for crossed fRIs
      dat_start <- paste0("_crossed", dat_gen, format(Sys.time(), "_%d_%b"))
    }else{
      if(!use_simple){
        #############
        # for one fRI
        dat_start <- paste0("_RI", dat_gen, format(Sys.time(), "_%d_%b"))  
      }else{
        ########################
        # for independent curves
        dat_start <- paste0("_simple", dat_gen, format(Sys.time(), "_%d_%b"))    
      }
    }
    
    ########################
    # iterate over levels of 
    # the grouping variables
    ########################
    for(I in Iseq){
      if(use_RI)
        Jseq <- NA  
      for(J in Jseq){
        
        ########################################################################
        # simulation run for
        # one iteration
        ########################################################################
        sim_r <- function(r){
          if(!file.exists(paste0("results/", f, "/res_I", I, "_J", J, "_r", r, dat_start, ".rds"))){
            cat("iteration: ", r,  "; I: ", I, "; J: ", J, "; ", f , "; time: ", format(Sys.time(), "%d %b %X"), "\n", sep = "")
            
            ###################
            # initialize output
            ###################
            output <- list()
            
            #######################
            # extract known 
            # components from truth
            #######################
            curve_info_true <- gen[[paste0("curve_info_true_I", I, "_J", J, "_r", r)]]
            
            if(!covariate){
              if(!use_RI){
                curve_info <- subset(curve_info_true, select = c(n_long, subject_long, 
                                                                 word_long, combi_long, number_long, t, y_vec))     
              }else{
                curve_info <- subset(curve_info_true, select = c(n_long, subject_long, 
                                                                 number_long, t, y_vec))   
              }
            }else{
              if(!use_RI){
                curve_info <- subset(curve_info_true, select = c(n_long, subject_long, 
                                                                 word_long, combi_long, number_long, t, y_vec))     
              }else{
                curve_info <- subset(curve_info_true, select = c(n_long, subject_long, 
                                                                 number_long, t, y_vec))       
              }
              for(k in 1:num_covariates){
                curve_info[, paste0("covariate.", k)] <- curve_info_true[[paste0("covariate.", k)]]
              }
            }
            
            ######################
            # remove unknown truth
            ######################
            curve_info_true <- NULL
            
            #######################
            # actual simulation run
            #######################
            output[["time_all"]] <- 
              system.time(output[["results"]] <- 
                            call_functions_fun(curve_info = curve_info, use_RI = use_RI, 
                                               use_simple = use_simple, method = method,  
                                               use_bam = use_bam, bs = bs, mp = mp, 
                                               para_estim = para_estim, para_estim_nc = para_estim_nc, 
                                               min_grid = min_grid, max_grid = max_grid, 
                                               my_grid = my_grid, bf_mean = bf_mean, 
                                               bf_covariates = bf_covariates, 
                                               m_mean = m_mean, covariate = covariate, 
                                               covariate_form = covariate_form, 
                                               num_covariates = num_covariates,  
                                               interaction = interaction, 
                                               which_interaction = which_interaction, 
                                               save_model_mean = save_model_mean, 
                                               bf_covs = bf_covs, m_covs = m_covs, 
                                               use_whole = use_whole, use_tri = use_tri, 
                                               use_tri_constr = use_tri_constr, 
                                               use_tri_constr_weights = use_tri_constr_weights, 
                                               np = np, var_level = var_level, 
                                               N_B = N_B, N_C = N_C, N_E = N_E, 
                                               use_famm = use_famm, use_bam_famm = use_bam_famm, 
                                               bs_int_famm = bs_int_famm, bs_y_famm = bs_y_famm, 
                                               para_estim_famm = para_estim_famm, 
                                               para_estim_famm_nc = para_estim_famm_nc, 
                                               save_model_famm = save_model_famm))
            
            #################
            # get sessionInfo
            #################
            sessionInfo <- sessionInfo()
            
            #############
            # save output
            #############
            cat("save output iteration: ", r,  "; I: ", I, "; J: ", J, "; ", f , "; time: ", format(Sys.time(), "%d %b %X"), "\n", sep = "")
            saveRDS(output, file = paste0("results/", f, "/res_I", I, "_J", J, "_r", r, dat_start, ".rds"))
            
            ############
            # save input
            ############
            cat("save input iteration: ", r,  "; I: ", I, "; J: ", J, "; ", f , "; time: ", format(Sys.time(), "%d %b %X"), "\n", sep = "")
            save(Iseq, Jseq, nrep, bf_mean, bf_covariates, bf_covs, method, m_mean, 
                 m_covs, bs, d_grid, var_level, N_B, N_C, N_E, use_bam, covariate, 
                 covariate_form, num_covariates, dat_gen, interaction, which_interaction, 
                 use_famm, mp, para_estim, para_estim_nc, use_tri, use_whole, 
                 use_tri_constr, use_tri_constr_weights, use_bam_famm, use_RI,
                 para_estim_famm, para_estim_famm_nc, save_model_mean, family, 
                 bs_int_famm, bs_y_famm, use_simple, np, min_grid, max_grid, 
                 my_grid, sessionInfo, save_model_famm, 
                 file = paste0("results/", f, "/input_I", I, "_J", J, "_r", r, dat_start, ".Rdata"))
          }else{
            warning(paste0("attention: file ", f, " I_", I, "_J", J, "_r", r, dat_start, " exists already"))
          }
        }
        
        ########################################################################
        # repeat for all iterations in parallel if on unix system
        # redo each iteration up to five times in case of abort and save
        # the information on the retries.
        ########################################################################
        reps_use <- 1:nrep
        do <- 1
        while(do < 5){	
          if(.Platform$OS.type == "unix"){         
            foreach(r = 1:length(reps_use)) %dopar%
              sim_r(reps_use[r])
          }else{
            foreach(r = 1:length(reps_use)) %do%
              sim_r(reps_use[r])
          }
          
          out <- numeric(nrep)
          for(r in 1:nrep){
            name_redo <- paste0("results/", f, "/res_I", I, "_J", J, "_r", r, dat_start, ".rds")
            out[r] <- file.exists(name_redo)  
          }          
          if(all(out == TRUE))
            break
          
          reps_use <- which(out == FALSE)
          warning(paste0("redo iterations ", reps_use ))
          save(reps_use, file = paste0("results/", f, "/not_worked_try", do, "_I", I, "_J", J, dat_start, ".Rdata"))
          do <- do + 1
        }
        warnings()
      }     
    }
    gen <- NULL
    gc()
  }
}
################################################################################
