####################################################################################################
# author: Jona Cederbaum
####################################################################################################
# description: specifies all input parameters for simulation for the scenario with indepedent curves
# and calls the main simulation function sim_face_fun.
####################################################################################################

###############
# load packages 
###############
require(face)
require("foreach")
if(.Platform$OS.type == "unix"){
  require("doMC")    # only possible under unix system
}

##################
# source functions
##################
source("sim_face.R")

###########################
# create folder for results
###########################
dir.create("results")


################################################################################
# specify input parameters 
################################################################################

#########################
# parallelization
# of simulation runs 
# (only for unix systems)
#########################
ncores <- 10           # number of cores used for parallelization of simulation runs
if(.Platform$OS.type == "unix"){
  registerDoMC(cores = ncores)
}


#######################
# which simulation data
#######################
dat_gen <- "_11_Mar"

######################
# number of iterations
######################
nrep_use <- NULL      # choose number of simulation iterations,
                      # which has to be smaller or equal the number of generated data sets
                      # NULL to use the number of generated data sets

############################################
# which distribution of the basis weights
# Gaussian basis weights: family <- "normal"
############################################
family <- "normal"
message(paste0("you are using family = ", family))
dir.create("results/normal")

###################
# general smoothing
###################
d_grid <- 100        # pre-specified grid length for equidistant grid on which the mean, the covariance surfaces, the eigenfunctions
                     # and the functional random effects are evaluated
min_grid <- 0        # minimum value of eqidistant grid (should approx. correspond to minimum value of time interval)
max_grid <- 1        # maximum value of eqidistant grid (should approx. correspond to maximum value of time interval)

################################
# mean and covariance estimation
################################
bf <- 10                 # dimension of basis (number of basis functions)
m <- list(c(2, 3))       # spline order and order of the penalty
                         # should be a list of length one
                         # e.g. c(2, 3) are cubic splines with third order difference penalty, see ?s for details
use_step_1_only <- FALSE # TRUE to estimate using FACE-STEP-1, i.e. only the first iteration is used

################
# FPC estimation
################
var_level <- 0.95        # pre-specified level of variance used for the choice of the number of the FPCs
                         # instead, a specific number of FPCs can be specified (see below)
calculate.scores <- TRUE # TRUE to calculate basis weights


################################################################################
# start simulation
################################################################################
sim_face_fun(bf = bf, m = m, d_grid = d_grid, var_level = var_level, dat_gen = dat_gen, 
             family = family, min_grid = min_grid, max_grid = max_grid, 
             calculate.scores = calculate.scores, use_step_1_only = use_step_1_only, 
             nrep_use = nrep_use)

################################################################################
# output parameters 
################################################################################
# output is a list which is saved in rds-format in a newly created results folder
# it consists of two elements:

## time_all: the total system.time() for calling function face.sparse()

## results: a list of all estimates further described in ?face.sparse


################################################################################
